#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>


/* TYPES OF DISTRIBUTIONS */
#define UNIFORM 0
#define EXPONENTIAL 1
#define NORMAL 2
#define DELTA 3

/* HISTORY-RELATED CONSTANTS */
#define MAX_JUNK_FRACTION 0.01
#define INIT_HISTORY_SIZE 10   /* make sure this isn't too big or things get really
sluggish because of so much memory getting allocated */


/* This is the developing version of a modification of markov10.c which allows selective bottlenecking.
 Markov 10.c allows death processes as well as birth.
 Additionally, capability to compete different strains has been added.  Also, a pre-equilibration run now exists,
 so that when strains start competing, their composition is at equilibrium.
 */

/* Simulates arbitrary markov birth, death, and jump processes in fluctuating environments */

/* Input: parameter file
 
 Parameters:
 COMPETE x1 ... xN        - if this flag is present, this is a competition, i.e. simulation ends when one strain wins, or SIM_LENGTH is reached
 x1 .. xN are the strain ID's of the N different phenotypes, allowing a single strain to possess multiple phenotypes.  
 These should be between 0 and N-1.
 
 PRE_EQUILIBRATE type cutoff par1 par2   
 - if this flag is present, the population will be propogated for a random time chosen from the distribution given by 
 type,cutoff,par1,par2 (see below).  
 If COMPETE is set, the pre-equilibration will be carried out separately for each strain, 
 after which they will be sampled equally and mixed.
 
 INITIAL_POP x1 .. xN     - gives the initial population sizes of the phenotypes
 
 RATE_MATS mat1 .. matM   - markov rate matrices for each of the M environments:
 
 Matrix elements Aij, for i != j, are the poisson rate of a jump from state i to state j.
 Matrix elements Aii are the birth (and/or death) rate of state i:
 if Aii > 0, the population of state i jumps from N -> N+1 with rate (Aii)*N
 if Aii < 0, it jumps from N -> N-1 with rate (Aii)*N
 if Aii = Aii,Bii (two numbers separated by a comma, w/ no space):
 
 ENV_DURS env_dur_file
 
 For each environment, the duration distributions file lists a line with following info:
 
 TYPE CUTOFF PAR1 PAR2
 
 where TYPE is one of the defined types of distributions (see above)
 CUTOFF is a number that is added to the random variable
 PAR1, PAR2 are two parameters defining the distribution (some need only 1)
 
 Histories:
 Histories are stored as lists of events, such that each event indicates the state that persisted
 (since the last event) until the time of the present event.  Thus, events are always taken to occur
 forward in time, and histories are output and stored forward in time.  When histories are output, 
 the start time of a history is given as the initial event.  The state of the initial event is therefore 
 always irrelevant, but is output for ease of use.
 
 In the output history file, each entry is separated by * *, and has the following form:
 Unique_ID Ancestor_ID    : 1 line
 state time               : n lines (each line is an event)
 
 The first line gives identifies the cell in question, and gives the ID of its progenitor.
 If Unique_ID = Ancestor_ID, this cell entry is a continuation of an entry from the previous file
 i.e., this cell survived the bottleneck.
 The value of state may be positive or negative.
 Negative values indicate a cell division, with -state giving the Unique_ID of the child cell; time gives the time of division
 Positive values that are < NState (which is the total number of phenotypic states) give the phenotypic state; time gives the time of switch.
 Positive values that are >= NState indicate a cell death; state-NState gives the phenotype at time of death; time gives the time of death.
 */
/* Technical notes:
 (1) Any routine that modifies population structure must update NPop
 */

/* random number generator */
gsl_rng *RNG;

/* simulation variables */
double **Bottle_Rates;
double **Distributions, **Env_Cume, Lyap, Avg_pop;
double ***Rate_Matrices, *Rates, **Birth_Probs, **Fitnesses, Tot_Rate, **Bij;
double *Prob_From, **Prob_To, Sim_Length, Print_Interval;
double Ttot, Tcycle, Tlast, Tperiod;
double Pre_Equil_Fixed_Dur, Pre_Equil_Rand_Dur;
long *Pop, Pop_Max, Pop_Sample, MAX_JUNK_POS, NCell_Sample, Round, NRuns, UniqueID, *Cell_ID;
long *Strain_Pop, NStrains, *Initial_Pop, NPop;
int  Cur_Env, Initial_Env;
int Seeder, NState, NEnv;
FILE *Environment_File, *History_File;

/* file names*/
char Output_Dir[1000], Env_Bij_Filename[1000], Rate_Mat_Token[1000];
char  Env_Durs_Filename[1000], Bottle_Rates_Filename[1000];

/* competition variables */
char Compete_Token[1000], Pre_Equilibrate_Token[1000], Initial_Pop_Token[1000];
short Competition, Pre_Equilibrate;
double Pre_Equil_Dist[5];


short Record_History, Record_All_Cells, *Strain_ID, Use_Initial_Pop;
long *Save_ID;
short *Save_State;

typedef struct {
	long state;
	double time;
} event;

event **History, *Cur_History;

long *Hist_Length, NHist, Max_NHist, *Alive_Cells;
long *Max_Hist_Length;

short *Cell_State;
long  **Cell_List;       /* for each state, gives a list of all alive cells in that state */
long  *Cell_List_Length; /* lengths of each list */
long *Max_Cell_List;     /* maximum allocated size (may change) */
long  **Junk_Pos;        /* a list of positions in Cell_List that are now junk (because a cell switched state or died) */
long *Junk_Pos_Length;   /* lengths of each list */

typedef struct {
	long cell;
	double time;
} branch;
branch *Ancestor;


int Randomint(int max) {
	/* the gsl routine can go up to an unsigned long int
     on the mac, an int, an unsigned int, and an unsigned long int 
     have the same physical size.  since unsigned int gives twice as many positive
     numbers as int, there is no problem using this routine as below.
	 */
	return ((int) gsl_rng_uniform_int(RNG, (unsigned long int) max));
}

long *ksub_random2 (long k, long n) {
	/*  Albert Nijenhuis and Herbert Wilf,
	 Combinatorial Algorithms,
	 Academic Press, 1978, second edition,
	 ISBN 0-12-519260-6. 
	 
	 original is in subset.C (in the markov/SOURCE directory)
	 */
	
	long available, candidate, have, need, *a;
	double r;
	
	a = (long *) calloc(k, sizeof(long));
	need = k;
	have = 0;
	available = n;
	candidate = -1;
	
	for ( ; ; ) {
		candidate = candidate + 1;
		
		r = gsl_rng_uniform(RNG);
		
		/*note: i changed this to be strictly less-than 
		 because take limiting case of 100 avail, and need 1.  then
		 you want to choose first one with prob 1/100, second 1/99 etc.
		 since r is drawn from [0,1), the strict less-than is needed here.  
		 original algorithm used an r drawn from (0,1). */
		
		if ( r * (double) available < (double) need ) {
			need = need - 1;
			a[have] = candidate;
			have = have + 1;
			if ( need <= 0 ) {
				break;
			}
		}
		available = available - 1;
		
	}
	return(a);
}


void Print_Matrix(double **matrix, int nrows, int ncols, FILE *outfile) {
	int i, j;
	
	for (i = 0; i < nrows; ++i){
		for (j = 0; j < ncols; ++j)
			fprintf(outfile,"%.3lf ", matrix[i][j]);
		fprintf(outfile,"\n");
	}
	
}

void Read_Rate_Matrix(char *mat_name, double ***matrix, int *nrows, int *ncols, double **birth_prob, double **fitness) {
	FILE *mat_file;
	char line[1001], field[100], *pos;
	double *death_rates;
	int i, j;
	
	/* reads a rate matrix, allocates memory, and returns matrix size  
     the diagonal is allowed to contain terms of the form x,y (no space)  
     in case there is a comma, x is interpreted as a birth rate, and y as a death rate  
     total rates, i.e. birth + abs(death), are returned in the matrix diagonal, and death rates  
     are converted into a birth probability, i.e. prob. that an event is a birth rather than a death,
     and returned as a separate vector */
	
	mat_file = fopen(mat_name, "r");
	fgets(line, 1000, mat_file);
	strtok(line, " \t");
	*ncols = 1;
	while (strtok(NULL, " \t")!=0)
		++(*ncols);
	fclose(mat_file);
	
	*nrows = 0;
	mat_file = fopen(mat_name, "r");
	while (fgets(line,1000,mat_file)!=NULL)
		++(*nrows);
	fclose(mat_file);
	
	*matrix = (double **) calloc(*nrows, sizeof(double *));
	for (i = 0; i < *nrows; ++i)
		(*matrix)[i] = (double *) calloc(*ncols, sizeof(double));
	death_rates = (double *) calloc(*nrows, sizeof(double));
	
	mat_file = fopen(mat_name, "r");
	for (i = 0; i < *nrows; ++i) {
		fgets(line,1000,mat_file); 
		strcpy(field,strtok(line," \t"));
		if (i==0) {
			pos = strchr(field,',');
			if (pos != NULL) {
				death_rates[i] = atof(&pos[1]);
				pos[0]='\0';
			}
		}
		(*matrix)[i][0] = atof(field);
		for (j = 1; j < *ncols; ++j) {
			strcpy(field,strtok(NULL," \t"));
			if (i==j) {
				pos = strchr(field,',');
				if (pos != NULL) {
					death_rates[i] = atof(&pos[1]);
					pos[0]='\0';
				}
			}
			(*matrix)[i][j] = atof(field);
		}
	}
	
	*birth_prob = (double *) calloc(*nrows, sizeof(double));
	*fitness = (double *) calloc(*nrows, sizeof(double));
	for (i = 0; i < *nrows; ++i) {
		if (death_rates[i] < 0)
			death_rates[i] = -death_rates[i];                    /* death rate is taken to be a positive number */
		(*fitness)[i] = (*matrix)[i][i] - death_rates[i];      /* fitness is used to calculate lyapunov exponent, so it is birth rate - death rate */
		(*matrix)[i][i] += death_rates[i];                     /* the Aii term is the total rate, so it is the sum of (the abs. vals of) birth & death rates */
		(*birth_prob)[i] = 1 - death_rates[i]/(*matrix)[i][i]; /* birth prob is computed from the magnitudes of the birth & death rates */
	}
	free(death_rates);
	
	fclose(mat_file);
}


void Read_Matrix(char *mat_name, double ***matrix, int *nrows, int *ncols) {
	FILE *mat_file;
	char line[1001];
	int i, j;
	
	/* reads an arbitrary size matrix, allocates memory, and returns matrix size */
	
	mat_file = fopen(mat_name, "r");
	fgets(line, 1000, mat_file);
	strtok(line, " \t");
	*ncols = 1;
	while (strtok(NULL, " \t")!=0)
		++(*ncols);
	fclose(mat_file);
	
	*nrows = 0;
	mat_file = fopen(mat_name, "r");
	while (fgets(line,1000,mat_file)!=NULL)
		++(*nrows);
	fclose(mat_file);
	
	*matrix = (double **) calloc(*nrows, sizeof(double *));
	for (i = 0; i < *nrows; ++i)
		(*matrix)[i] = (double *) calloc(*ncols, sizeof(double));
	
	mat_file = fopen(mat_name, "r");
	for (i = 0; i < *nrows; ++i) {
		fgets(line,1000,mat_file); 
		(*matrix)[i][0] = atof(strtok(line," \t"));
		for (j = 1; j < *ncols; ++j)
			(*matrix)[i][j] = atof(strtok(NULL," \t"));
	}
	fclose(mat_file);
}

void Check_Matrix(double **matrix, int msize) {
	int i, j;
	
	/* check that jump rates are all positive */
	
	for (i = 0; i < msize; ++i)
		for (j = 0; j < msize; ++j)
			if (i != j)
				if (matrix[i][j] < 0.0)
					printf("rate %d-%d is negative!\n", i, j);
	
}


void Calc_Rates(double **matrix, int msize, long *pop, double *rates, double *tot_rate) {
	int i, j;
	
	*tot_rate = 0;
	for (i = 0; i < msize; ++i) {
//		rates[i] = 0;
//		for (j = 0; j < msize; ++j)
//			rates[i] += matrix[i][j];
        
        /* use only the diagonal, since Rates gives the division rates in this code */
        rates[i] = matrix[i][i];
		*tot_rate += pop[i]*rates[i];
	}
	
}

void Update(double **rate_matrix, int msize, long *pop, double *rates, double *tot_rate){
	int i;
	
	*tot_rate = 0;
	for (i = 0; i < msize; ++i)
		*tot_rate += pop[i]*rates[i];
	
	Prob_From[0] = pop[0]*rates[0]/(*tot_rate);
	for (i = 1; i < msize-1; ++i) 
		Prob_From[i] = Prob_From[i-1] + pop[i]*rates[i]/(*tot_rate); 
	
}

void Setup_Probs(double **rate_matrix, int msize, long *pop, double *rates, double tot_rate) {
	int i, j;
    double tot_prob;
	
	/* Given that an event occured,
     Prob_From[i]  is the probability that the jump occured from any state <= i 
     Prob_To[i][j] is the probability that (given the "from" state i),
	 the jump occured to any state <= j 
     (so these are actually the cumulative distributions)
     The "from" probabilities depend on the population vector
     while the "to" probabilities do not.
	 */
	
	Prob_From[0] = pop[0]*rates[0]/tot_rate;
	for (i = 1; i < msize-1; ++i) 
		Prob_From[i] = Prob_From[i-1] + pop[i]*rates[i]/tot_rate; 
	
	for (i = 0; i < msize; ++i) {
        tot_prob = 0;
        for (j = 0; j < msize; ++j)
            if (i!=j) {
                tot_prob += rate_matrix[i][j];
                Prob_To[i][j] = rate_matrix[i][j];
            }
        if ((tot_prob < 0) || (tot_prob > 1)){
            printf("Bad probability in Setup_Probs\n");
            exit(-1);
        }
        Prob_To[i][i] = 1 - tot_prob;
		for (j = 1; j < msize-1; ++j)
			Prob_To[i][j] += Prob_To[i][j-1];
	}
}

int Discrete_Sample(double *cume, int msize) {
	double myran;
	int i;
	
	/* samples from a discrete distribution, with msize bins
     with cumulative distribution given by cume[i] */
	
	myran = gsl_rng_uniform(RNG);
	for (i = 0; (i < msize-1) && (myran >= cume[i]); ++i);
	
	return(i);
}



void Print_State(void) {
	int i;
	
	printf("%lf ", Ttot);
	for (i = 0; i < NState; ++i)
		printf("%ld ", Pop[i]);
	printf("%lf ", Lyap);
	printf("%0.1lf\t", 1/Avg_pop);
	
	if (Competition)
		for (i = 0; i < NStrains; ++i)
			printf("\t%ld",Strain_Pop[i]);
	printf("\t%ld\n", NPop);
    fflush(stdout);
	
}



long Get_Pop_Size(void) {
	long npop=0;
	int i;
	
	for (i=0; i < NState; ++i)
		npop+=Pop[i];
	return(npop);
}



void Resample_Separately(void) {
    
	/* For Competitions only, resamples population vector, on a strain-by-strain basis, 
       down to (roughly) NCell_Sample cells for each strain.
       Returns the new population size (which is exactly new_size unless noisey sampling is used) */
	int i;
	long ntot=0;
    double *exp_size, *factor;
	
    exp_size = (double *) calloc(NStrains, sizeof(double));
    factor = (double *) calloc(NStrains, sizeof(double));
	
    for (i = 0; i < NState; ++i) {
        exp_size[Strain_ID[i]]+=Bottle_Rates[Cur_Env][i]*Pop[i];
    }
    
    for (i=0; i<NStrains; ++i){
        factor[i] = 1;
        if (exp_size[i] > NCell_Sample)
            factor[i] = NCell_Sample/exp_size[i];
    }
	
	for(i = 0; i < NState; ++i) {
		/*Pop[i] = floor((Pop[i]/(double)(ncell_strain[Strain_ID[i]]))*NCell_Sample + 0.5);*/
		Pop[i] = gsl_ran_poisson(RNG,  Bottle_Rates[Cur_Env][i]*Pop[i]*factor[Strain_ID[i]]);
		ntot += Pop[i];
	}
    
    for (i = 0; i < NStrains; ++i)
        Strain_Pop[i] = 0;
    for (i = 0; i < NState; ++i)
        Strain_Pop[Strain_ID[i]] += Pop[i];
    
    free(factor);
    free(exp_size);
	NPop = ntot;
}

void Resample(void) {
	int i;
	long ntot=0;
    double exp_size=0.0, factor=1.0;

    /* check the expected size of population after bottleneck rates applied*/
	for (i = 0; i < NState; ++i) {
        exp_size += Bottle_Rates[Cur_Env][i]*Pop[i];
    }
    
    /* if expected size is larger than bottleneck size NCell_Sample, rescale sampling by factor */
    if (exp_size > NCell_Sample) {
        factor = NCell_Sample/exp_size;
    }
    
    /* sample down with Poisson statistics on bottleneck rates */
    for (i = 0; i < NState; ++i) {
        /*Pop[i] = floor((Pop[i]/(double)NPop)*NCell_Sample + 0.5);*/
		Pop[i] = gsl_ran_poisson(RNG, Bottle_Rates[Cur_Env][i]*Pop[i]*factor);
		ntot += Pop[i];
	}
    
	if (Competition) {
		for (i = 0; i < NStrains; ++i) 
			Strain_Pop[i] = 0;
		for (i = 0; i < NState; ++i)
			Strain_Pop[Strain_ID[i]] += Pop[i];
	}
	NPop = ntot;
}


double Sample_Distribution(double *dist) {
	double value;
	
	switch ((int) dist[0]) {
			
		case UNIFORM: /* uniformly distributed from [0,dist[2]) */
			value = dist[1] + dist[2]*gsl_rng_uniform(RNG); 
			break;
			
		case EXPONENTIAL: /* exponentially distributed with mean dist[2] */
			value = dist[1] + gsl_ran_exponential(RNG, dist[2]);
			break;
			
		case NORMAL: /* normally distributed with mean dist[2] and stddev dist[3] */
			value = dist[1] + dist[2] + gsl_ran_gaussian(RNG, dist[3]);
			break;
			
		case DELTA:
			value = dist[1] + dist[2];
			break;
			
	}
	
	return(value);
}




void Clean_Junk (short state) {
	long i, j=0;
	
	for (i = 0; i < Cell_List_Length[state]; ++i)
		if (Cell_List[state][i] >= 0)
			Cell_List[state][j++] = Cell_List[state][i];
	
	Cell_List_Length[state] = j;
	Junk_Pos_Length[state] = 0;
}



long *Get_Alive_Cells(long *ncells) {
	/* returns the list of alive cells */
	int state;
	long  i, j=0;
	long *alive_cells;
	
	*ncells = 0;
	for (state=0; state < NState; ++state) {
		if (Junk_Pos_Length[state] > 0)
			Clean_Junk(state);
		*ncells += Cell_List_Length[state];
	}
	
	alive_cells = calloc(*ncells, sizeof(long));
	for (state=0; state < NState; ++state)
		for (i = 0; i < Cell_List_Length[state]; ++i)
			alive_cells[j++] = Cell_List[state][i];
	
	return(alive_cells);
}



void Finalize_History (long the_cell, double final_time) {
	int event_num;
	
	/* check max size of the given history */
	if (Max_Hist_Length[the_cell] == Hist_Length[the_cell]) {
		Max_Hist_Length[the_cell]++;
		History[the_cell] = (event *) realloc(History[the_cell], Max_Hist_Length[the_cell]*sizeof(event));
	}
	
	/* add the current state of the cell at time final_time to the history */
	event_num = Hist_Length[the_cell];
	History[the_cell][event_num].state = Cell_State[the_cell];
	History[the_cell][event_num].time = final_time;
	Hist_Length[the_cell]++;
}



void Finalize_Histories (double final_time) {
	/* this adds the final timestamp to all cells that are alive at the present time */
	/* the trick here is not to add the stamp to dead cells, since they already got their final stamp */
	int j;
	long nalive;
	
	Alive_Cells = Get_Alive_Cells(&nalive);
	
	for (j = 0; j < nalive; ++j)
		Finalize_History(Alive_Cells[j], final_time);
}



void Check_Size (long cur_size, long *max_size, long block_size, void **the_ptr) {
	if (*max_size == cur_size) {
		*max_size += INIT_HISTORY_SIZE;    /* used to be doubled... this is more conservative */
		*the_ptr =  realloc(*the_ptr, (*max_size)*block_size);
		if (*the_ptr==NULL) {
			printf("\nOut of Memory\n");
			exit(0);
		}
	}
}

void Realloc_Histories () {
	long i;
	
	Max_NHist += Pop_Max;
	
	History = (event **) realloc(History, Max_NHist*sizeof(event *));
	Hist_Length = (long *) realloc(Hist_Length, Max_NHist*sizeof(long));
	Max_Hist_Length = (long *) realloc(Max_Hist_Length, Max_NHist*sizeof(long));         
	Cell_State = (short *) realloc(Cell_State, Max_NHist*sizeof(short));           
	Cell_List = (long **) realloc(Cell_List, NState*sizeof (long *));             
	Max_Cell_List = (long *) realloc(Max_Cell_List, NState*sizeof (long));      
	Ancestor = (branch *) realloc(Ancestor, Max_NHist*sizeof(branch));     
	Cell_ID = (long *) realloc(Cell_ID, Max_NHist*sizeof(long));               
}


void Update_History (short from_state, short to_state, short birth, double time_stamp) {
	long the_cell, i, j;
	int event_num, list_pos;
	
	/* check that max allocated size of cell_list[from_state] is enough */
	Check_Size(Pop[to_state], &Max_Cell_List[to_state], sizeof(long), (void **) &Cell_List[to_state]);
	
	/* remove junk positions if there are too many */
	if ((Junk_Pos_Length[from_state] >= Cell_List_Length[from_state]*MAX_JUNK_FRACTION) || (Junk_Pos_Length[from_state] == MAX_JUNK_POS))
		Clean_Junk(from_state);
	
	/* choose a cell of the given type randomly - iterate until a non-junk position is found */
	the_cell = -1;
	while (the_cell == -1) {
		list_pos = Randomint(Cell_List_Length[from_state]);
		the_cell = Cell_List[from_state][list_pos];
	}
	
	/* check max size of the given history */
	Check_Size(Hist_Length[the_cell], &Max_Hist_Length[the_cell], sizeof(event), (void **) &History[the_cell]);
	
    if (birth) {
        if (NHist == Max_NHist)
            Realloc_Histories();
        
        /* allocate memory for new cell's history */
        History[NHist] = (event *) calloc(INIT_HISTORY_SIZE, sizeof(event));
        if (History[NHist]==NULL) {
            printf("\nOut of Memory\n");
            exit(0);
        }
        Hist_Length[NHist] = 0;
        Max_Hist_Length[NHist] = INIT_HISTORY_SIZE;
        Ancestor[NHist].cell = the_cell;
        Ancestor[NHist].time = time_stamp;
        
        /* record the split in the ancestral lineage */
        event_num = Hist_Length[the_cell];
        History[the_cell][event_num].state = -UniqueID;   /* -1 marks a split event */
        History[the_cell][event_num].time = time_stamp;
        Hist_Length[the_cell]++;
        
        /* assign unique ID to the new cell */
        Cell_ID[NHist] = UniqueID++;
        
        /* update the list of cell states */
        Cell_State[NHist] = to_state; /* note in this model switching occurs at birth */
        
        /* add the new cell to the cell_list[from_state] */
        if (Junk_Pos_Length[to_state] > 0)
        /* if there are junk positions available, use these first */
            Cell_List[to_state][Junk_Pos[to_state][--Junk_Pos_Length[to_state]]]=NHist;
        else
        /* no junk positions, so add a new element to cell_list */
            Cell_List[to_state][Cell_List_Length[to_state]++]=NHist;
    }
    
    else /* death */ {
        /* remove the cell from Cell_List[from_state] */
        Cell_List[from_state][list_pos] = -1;   /* turns given position to a junk position */
        Junk_Pos[from_state][Junk_Pos_Length[from_state]++] = list_pos;  /* adds the position to list of junk positions */
        
        /* record time of death in its history */
        Cell_State[the_cell] += NState; /* this marks it as a death event */
        Finalize_History(the_cell, time_stamp);
    }
    
}


short No_Winner (void) {
	/* this routine tests whether one strain has prevailed over all others */
	short i, no_winner=1;
	
	for (i = 0; i < NStrains; ++i) 
		if (Strain_Pop[i]==NPop)
			no_winner=0;
	return(no_winner);
}


short Who_Won (void) {
	int i;
	
	if (No_Winner())
		return(-1);
    for (i = 0; i < NState; ++i)
        if (Pop[i] > 0)
            break;
    return(Strain_ID[i]);
}


void Simulate(void) {
	double tnext, **rate_matrix, *fitness, *birth_prob, dlyap;
	int i, j;
	short period_over, birth;
	
	/*  Ttot - the total simulation time
        Tcycle - elapsed time in the current cycle
        Tlast - time of last printout  */
    
    period_over = 0;
    
	NHist = NPop; /* this is the total number of histories; it may be greater than total population size if cells can die */
	
	/* rate_matrix contains the current rate matrix */
	rate_matrix = Rate_Matrices[Cur_Env];
	fitness = Fitnesses[Cur_Env];
	birth_prob = Birth_Probs[Cur_Env];
    
	while ((NPop > 0) && (NPop < Pop_Max) && (Tcycle < Tperiod) && (Ttot < Sim_Length) && (!Competition || No_Winner())) {
		/* calculate the time of the next event from the total rate */
		tnext = gsl_ran_exponential(RNG, 1/Tot_Rate);
        
		/* if tnext brings the system out of the current period, the event did not happen so */
        if(Tcycle+tnext>Tperiod) {
            /* increment clock to end of the period */
            tnext = Tperiod - Tcycle;
            period_over = 1; /* signal to break loop */
        }
	
        /* if tnext brings the system beyond simulation time, the event did not happen so */
		if ((Ttot+tnext>Sim_Length)) {
			tnext = Sim_Length - Ttot; /* increment clock to end of simulations */
            period_over = 1; /* signal to break loop */
			if (Tcycle > 0.0) {
				fprintf(Environment_File, "%d %.4lf\n", Cur_Env, Ttot+tnext); /* final record of environment */
				
			}
		}
		
		dlyap = 0;
		for (i = 0; i < NState; ++i)
			dlyap += ((double)Pop[i]/NPop)*fitness[i];
		Lyap = (Lyap*Ttot + dlyap*tnext)/(Ttot + tnext);
		Avg_pop = (Avg_pop*Ttot + (1/(double)NPop)*tnext)/(Ttot + tnext);
		
		/* increment the clocks */
		Ttot += tnext;
		Tcycle+=tnext;
        
        /* if event happened after current period or simulation end, nothing happened, so break */
        if (period_over) break; /* otherwise, continue to determine event and update rates */
        
		/* sample from the cumulative distributions to find the leaving and arriving states */
		do {
			i = Discrete_Sample(Prob_From, NState);
		} while ((Rates[i] == 0.0) || (Pop[i]==0));  /* should never have to go through twice! see if this ever happens*/
		do {
			j = Discrete_Sample(Prob_To[i], NState); 
		} while (Prob_To[i][j] == 0.0); /*same here */
		
        /* here we need to decide if event was a birth or death */
        birth = ((birth_prob[i] == 1) || (gsl_rng_uniform(RNG) < birth_prob[i]));
        
        /* update history */
        if (Record_History) Update_History(i, j, birth, Ttot);
        
        /* Update population counters */
        if (birth) {
            Pop[j]++;
            if (Competition)
                Strain_Pop[Strain_ID[j]]++;
            NHist++;
            NPop++;
        }
        else {
            Pop[i]--;
            if (Competition)
                Strain_Pop[Strain_ID[i]]--;
            NPop--;
        }
       
		/* update the rates and probabilities as a result of the population changes */
		Update(rate_matrix, NState, Pop, Rates, &Tot_Rate);
		
		/* print population vector */
		if (Ttot - Tlast > Print_Interval) { 
			Print_State();
			Tlast = Ttot; 
		}
	}
    
    /* if loop ended because NPop exceeded limit, this is equivalent to end of a cycle,
       so increment the clock to end of cycle */
    
    
//    if (NPop >= Pop_Max) {
//        /* print state since the clock jump may have gone beyond Print_Interval */
//        Print_State();
//        
//        /* increment clock to end of the period */
//        Ttot += Tperiod - Tcycle;
//    
//        Tlast = Ttot;
//    }
//    
	
	if (Record_History) Finalize_Histories(Ttot);
}

event *Reconstruct_History(long the_cell, long *final_length, long *ancestor_cell) {
	long cur_cell, cur_len, hist_len=0, j, start_point, max_hist_size;
	double last_time;
	event *the_hist, *final_hist;
	
	/* this routine follows the history of a cell backwards, passing through all the ancestors */
	/* and pulls out the complete history backwards, stored in the_hist */
	
	max_hist_size = INIT_HISTORY_SIZE; 
	/* max_hist_size = 1000; */
	the_hist = (event *) calloc(max_hist_size, sizeof(event));
	cur_cell = the_cell;
	last_time = Ttot + 1;
	while (cur_cell != -1) {
		cur_len = Hist_Length[cur_cell];
		j = cur_len-1;
		while ((j >= 0) && (History[cur_cell][j].time >= last_time))
			--j;
		start_point = j;
		if (start_point != -1) {
			the_hist = (event *) realloc(the_hist, (hist_len+start_point+1)*sizeof(event)); 
			for (j = start_point; j >= 0; --j)
				the_hist[hist_len++] = History[cur_cell][j];
		}
		last_time = Ancestor[cur_cell].time;
		*ancestor_cell = cur_cell;
		cur_cell = Ancestor[cur_cell].cell;
	}
	
	/* this is just a simple reversal of the_hist, since events were stored backwards from Ttot to zero */
	final_hist = (event *) calloc(hist_len, sizeof(event));
	for (j = 0; j < hist_len; ++j)
		final_hist[j] = the_hist[hist_len-1-j];
	free(the_hist);
	*final_length = hist_len;
	return(final_hist);
}



void Init_History () {
	long i;
	
	Max_NHist = Pop_Max;
	
	History = (event **) calloc(Max_NHist, sizeof(event *));            /* for cell i, History[i] records the list of events in its life-history */
	Hist_Length = (long *) calloc(Max_NHist, sizeof(long));             /* for cell i, Hist_Length[i] gives the number of events in its history */
	Max_Hist_Length = (long *) calloc(Max_NHist, sizeof(long));         /* the allocated size (in number of events )for the history of cell i */
	
	for (i = 0; i < NCell_Sample; ++i) {
		Max_Hist_Length[i] = INIT_HISTORY_SIZE;
		History[i] = (event *) calloc(INIT_HISTORY_SIZE, sizeof(event));  /* allocates space for the history of the initial cells */
	}
	Cell_State = (short *) calloc(Max_NHist, sizeof(short));            /* a list of the current state of each cell */
	Cell_List = (long **) calloc(NState, sizeof (long *));              /* for each type of cell, a list of all cells of that type */
	Max_Cell_List = (long *) calloc(NState, sizeof (long));             /* the maximum allocated size of Cell_List[i] */
	
	Ancestor = (branch *) calloc(Max_NHist, sizeof(branch));            /* for cell i, Ancestor[i] gives the time of birth, and ancestor cell */
	
	Cell_ID = (long *) calloc(Max_NHist, sizeof(long));                 /* for each cell, gives the unique ID number */
	
	MAX_JUNK_POS = (int) (Max_NHist*MAX_JUNK_FRACTION);                 /* the maximum amount of junk that can be stored for any cell type */
	/* MAX_JUNK_FRACTION is the fraction of positions of Cell_List[i] that */
	/* will be tolerated as junk */
	
	if (MAX_JUNK_POS == 0)
		MAX_JUNK_POS = 1;
	
	Junk_Pos = (long **) calloc(NState, sizeof (long *));        
    /* for each type (i) of cell, Junk_Pos[i] is a list of all positions of Cell_List[i] that are junk, rather than cells */
    /* the idea is that instead of shifting Cell_List[i] each time a cell of type i changes state, I record a -1 in the given position */
    /* when too many -1's accumulate, I clean up Cell_List[i], and reset Junk_Pos[i] so there is no junk */
	
	for (i = 0; i < NState; ++i) {
		Cell_List[i] = (long *) calloc(Pop_Sample, sizeof(long));
		Junk_Pos[i] = (long *) calloc(MAX_JUNK_POS, sizeof(long));
		Max_Cell_List[i] = Pop_Sample;
	}
	
	Junk_Pos_Length = (long *) calloc(NState, sizeof (long));    /* the total number of junk positions in Cell_List[i] */
	Cell_List_Length = (long *) calloc(NState, sizeof (long));   /* the total length of Cell_List[i] including junk positions */
	/* this number is used when drawing a random element of Cell_List[i] */
	/* if there is any junk, this number is strictly greater than Pop[i] */
}



void Free_History () {
	long i;
	
	free(Max_Cell_List);
	free(Cell_List_Length);
	free(Junk_Pos_Length);
	for (i = 0; i < NState; ++i) {
		free(Cell_List[i]);
		free(Junk_Pos[i]);
	}
	free(Cell_List);
	free(Junk_Pos);
	free(Cell_ID);
	free(Ancestor);
	free(Cell_State);
	for (i = 0; i < NHist; ++i)
		free(History[i]);
	free(History);
	free(Max_Hist_Length);
	free(Hist_Length);
}



void Output_Histories (long *save_cell, long ncells, long round) {
	FILE *out_file;
	char file_name[100];
	long i, save_hist_length, ancestor_cell, j;
	event *save_history;
	
	fprintf(History_File, "file %ld\n", round);
	if (!Record_All_Cells) 
		for (i = 0; i < ncells; ++i) {
			save_history = Reconstruct_History(save_cell[i], &save_hist_length, &ancestor_cell); 
			fprintf(History_File, "%ld %ld\n", i, ancestor_cell); 
			for (j = 0; j < save_hist_length; ++j) {
				if (save_history[j].state < 0) save_history[j].state = -1;
				fprintf(History_File, "%ld %.4lf\n", save_history[j].state, save_history[j].time); 
			}
			fprintf(History_File, "* *\n"); 
			free(save_history);
		}
	else {
		for (i = 0; i < NHist; ++i) {
			ancestor_cell = Ancestor[i].cell;
			if (ancestor_cell == -1)
				ancestor_cell = Cell_ID[i];
			else
				ancestor_cell = Cell_ID[ancestor_cell];
			fprintf(History_File, "%ld %ld\n", Cell_ID[i], ancestor_cell);
			for (j = 0; j < Hist_Length[i]; ++j)
				fprintf(History_File, "%ld %.4lf\n", History[i][j].state, History[i][j].time);  
			fprintf(History_File, "* *\n"); 
		}
	}
}



void Output_Parameters(void) {
	FILE *parameter_file;
	long i, j;
	char param_file_name[1000];
	
	sprintf(param_file_name, "%s/parameters", Output_Dir);
	parameter_file = fopen(param_file_name, "w");
	fprintf(parameter_file, "pop_sample\t\t%ld\n", Pop_Sample);
	fprintf(parameter_file, "pop_max\t\t\t%ld\n", Pop_Max);
	fprintf(parameter_file, "sim_length\t\t%.3lf\n", Sim_Length);
	fprintf(parameter_file, "true_sim_length\t\t%.3lf\n", Ttot);
	fprintf(parameter_file, "nrounds\t\t\t%ld\n", Round-1);
	fprintf(parameter_file, "print_interval\t\t%.3lf\n", Print_Interval);
	fprintf(parameter_file, "nstate\t\t\t%d\n", NState);
	fprintf(parameter_file, "nenvs\t\t\t%d\n", NEnv);
	fprintf(parameter_file, "bij_matrix\n");
	Print_Matrix(Bij, NEnv, NEnv, parameter_file);
	fprintf(parameter_file, "env_durations_matrix\n");
	Print_Matrix(Distributions, NEnv, 4, parameter_file);
	fprintf(parameter_file, "rate_matrices\n");
	for (i = 0; i < NEnv; ++i)
		Print_Matrix(Rate_Matrices[i], NState, NState, parameter_file);
	fprintf(parameter_file, "birth_probs\n");
	for (i = 0; i < NEnv; ++i) {
		for (j = 0; j < NState; ++j)
			fprintf(parameter_file, "%.3f ", Birth_Probs[i][j]);
		fprintf(parameter_file, "\n");
	}
	fclose(parameter_file);
}



int Count_Lines(char *file_name) {
	FILE *the_file;
	int nlines;
	char *line;
	
	line = calloc(10000001, sizeof(char));
	nlines = 0;
	the_file = fopen(file_name, "r");
	while (fgets(line,10000000,the_file)!=NULL)
		++(nlines);
	fclose(the_file);
	free(line);
	
	return(nlines);
}



void Assign_Param(char *param) {
	if (!strcmp(param, "POP_SAMPLE"))
		Pop_Sample = (long) atof(strtok(NULL," \t\n"));
	else if (!strcmp(param, "POP_MAX"))
		Pop_Max = (long) atof(strtok(NULL," \t\n"));
	else if (!strcmp(param, "SIM_LENGTH"))
		Sim_Length = (double) atof(strtok(NULL," \t\n"));
	else if (!strcmp(param, "NRUNS"))
		NRuns = (long) atof(strtok(NULL," \t\n"));
	else if (!strcmp(param, "ENV_BIJ"))
		strcpy(Env_Bij_Filename, strtok(NULL," \t\n"));
	else if (!strcmp(param, "ENV_DURS"))
        strcpy(Env_Durs_Filename, strtok(NULL," \t\n"));
    else if (!strcmp(param, "BOTTLE_RATES"))
        strcpy(Bottle_Rates_Filename, strtok(NULL," \t\n"));
	else if (!strcmp(param, "RATE_MATS"))
		strcpy(Rate_Mat_Token, strtok(NULL,""));
	else if (!strcmp(param, "COMPETE"))
		strcpy(Compete_Token, strtok(NULL,""));
	else if (!strcmp(param, "PRE_EQUILIBRATE"))
		strcpy(Pre_Equilibrate_Token, strtok(NULL,""));
	else if (!strcmp(param, "INITIAL_ENV"))
		Initial_Env = (int) atof(strtok(NULL," \t\n"));
	else if (!strcmp(param, "INITIAL_POP")) {
		Use_Initial_Pop = 1;
		strcpy(Initial_Pop_Token, strtok(NULL,""));
	}
	else if (!strcmp(param, "OUTPUT_DIR"))
		strcpy(Output_Dir, strtok(NULL," \t\n"));
	else if (!strcmp(param, "PRINT_INTERVAL"))
		Print_Interval = (double) atof(strtok(NULL," \t\n")); 
	else if (!strcmp(param, "RECORD_HISTORY"))
		Record_History = (short) atof(strtok(NULL," \t\n"));
	else if (!strcmp(param, "RECORD_ALL_CELLS"))
		Record_All_Cells = (short) atof(strtok(NULL," \t\n"));
}



void Inits(int argc, char *argv[]) {
	char env_file_name[1000], line[1000], param[1000];
	char **rate_mat_names;
	int nrows, ncols, i, j;
    unsigned long int seed;
    FILE *param_file;
	
    /* init random number generator */
    RNG = gsl_rng_alloc (gsl_rng_ranlxd2);
    // RNG = gsl_rng_alloc (gsl_rng_mt19937);
    seed = time(NULL);
    if (argc > 2) {
        seed = (unsigned long int) atof(argv[2]);
    }
    gsl_rng_set(RNG, seed);
    printf("#random number seed: %ld\n", seed);
    
	/* A couple globals that need to be initialized before parameters are read */
	Compete_Token[0] = '\0';
	Pre_Equilibrate_Token[0] = '\0';
	Use_Initial_Pop = 0;
	Initial_Env = -1;
	
	/* read parameters */
	param_file = fopen(argv[1], "r");
	while (fgets(line,1000,param_file)!=NULL) {
		if ((line[0] != ' ') && (line[0] != '-') && (line[0] != '\n')) {
			strcpy(param, strtok(line, " \t\n"));
			Assign_Param(param);
		}
	}
	fclose(param_file);
	
	/* maximum number of histories to allocate - keep this here b/c it could change at Realloc_Histories */
	Max_NHist = Pop_Max;
	
	/* number of environments */
	NEnv = Count_Lines(Env_Durs_Filename);
	
	/* number of internal states */
	rate_mat_names = (char **) calloc(NEnv, sizeof(char *));
	for (i=0; i < NEnv; ++i) {
		rate_mat_names[i] = (char *) calloc(1000, sizeof(char));
		if (i==0)
			strcpy(rate_mat_names[i], strtok(Rate_Mat_Token, " \t\n"));
		else 
			strcpy(rate_mat_names[i], strtok(NULL, " \t\n"));
	}
	NState = Count_Lines(rate_mat_names[0]);
	Pop = (long *) calloc(NState, sizeof(long));
	
	/* is there an initial pop vector supplied? */
	if (Use_Initial_Pop) {
		Initial_Pop = (long *) calloc(NState, sizeof(long));
		Initial_Pop[0] = (long) atof(strtok(Initial_Pop_Token, " \t\n"));
		for (i=1; i < NState; ++i) 
			Initial_Pop[i] = (long) atof(strtok(NULL, " \t\n"));
	}
	
	/* is this a competition? */
	Competition = 0;
	Pre_Equilibrate = 0;
	NStrains = 1;
	if (strlen(Compete_Token) > 0) {
		Strain_ID = (short *) calloc(NState, sizeof(short));
		Strain_Pop = (long *) calloc(NState, sizeof(long));
		Strain_ID[0] = (short) atof(strtok(Compete_Token, " \t\n"));   /* maps states onto strains, i.e. which types belong to which strains */
		for (i=1; i < NState; ++i)
			Strain_ID[i] = (short) atof(strtok(NULL, " \t\n"));
		for (i=0; i < NState; ++i)     
			if (Strain_ID[i] >= NStrains)
				NStrains = Strain_ID[i]+1;   /* the maximum strain ID is taken to be the number of strains, i.e. strain ID's should be numbers between 0 and NStrains-1. */
		if (NStrains > 1)
			Competition = 1;
	}
	
	/* is there a pre-equilibration step? */
	if (strlen(Pre_Equilibrate_Token) > 0) { 
		Pre_Equil_Dist[0] = atof(strtok(Pre_Equilibrate_Token, " \t\n"));
		Pre_Equil_Dist[1] = atof(strtok(NULL, " \t\n"));
		Pre_Equil_Dist[2] = atof(strtok(NULL, " \t\n"));
		Pre_Equil_Dist[3] = atof(strtok(NULL, " \t\n"));
		Pre_Equilibrate = 1;
	}
	
	/* read bij matrix for environmental transitions */
	Read_Matrix(Env_Bij_Filename, &Bij, &nrows, &ncols);
	if ((nrows != NEnv) || (ncols != NEnv))
		fprintf(stderr,"Size of Bij matrix does not match number of environments.\n");
	
	/* calculate cumulative distributions for environment jumps */
	Env_Cume = (double **) calloc(NEnv, sizeof (double *));
	for (i = 0; i < NEnv; ++i) {
		Env_Cume[i] = (double *) calloc(NEnv, sizeof(double));
		Env_Cume[i][0] = Bij[0][i];
		for (j = 1; j < NEnv - 1; ++j)
			Env_Cume[i][j] = Env_Cume[i][j-1] + Bij[j][i]; 
	}
	
    /* read bottleneck rates for each environment (row) */
    Read_Matrix(Bottle_Rates_Filename, &Bottle_Rates, &nrows, &ncols);
    if (nrows != NEnv)
        fprintf(stderr,"Number of bottleneck distributions does not match number of environments.\n");
   
    /* read distributions of environmental durations */
	Read_Matrix(Env_Durs_Filename, &Distributions, &nrows, &ncols);
	if (nrows != NEnv)
		fprintf(stderr,"Number of distributions does not match number of environments.\n");
	
	/* for each env, its rate matrix, with diagonal entries giving the total rate of births and deaths, i.e. birthrate+deathrate */
	Rate_Matrices = (double ***) calloc(NEnv, sizeof(double **));
	
	/* for each env and phenotype, prob that birth occurs, i.e. birthrate/(birthrate+deathrate) */
	Birth_Probs = (double **) calloc(NEnv, sizeof(double *));  
	
	/* for each env and phenotype, net growth rate, i.e. birthrate - deathrate */
	Fitnesses = (double **) calloc(NEnv, sizeof(double *));        
	
	/* read rate matrices for each environment */
	for (i = 0; i < NEnv; ++i) {
		Read_Rate_Matrix(rate_mat_names[i], &Rate_Matrices[i], &nrows, &ncols, &Birth_Probs[i], &Fitnesses[i]);
		Check_Matrix(Rate_Matrices[i], NState);
	}
	
	/* these are the total rates for each state */
	Rates = (double *) calloc(NState, sizeof(double));
	
	/* initialize probability arrays, to and from */
	Prob_From = (double *) calloc(NState, sizeof(double));
	Prob_To = (double **) calloc(NState, sizeof(double *));
	for (i = 0; i < NState; ++i)
		Prob_To[i] = (double *) calloc(NState, sizeof(double));
	
	for (i=0; i < NEnv; ++i)
		free(rate_mat_names[i]);
	free(rate_mat_names);
	
	
	/* this array keeps track of the initial states of Pop_Sample cells that found the next round */
	Save_State = (short *) calloc(Pop_Sample, sizeof(short));
	Save_ID = (long *) calloc(Pop_Sample, sizeof(long));
}



void Read_History(char *file_name, event **history, long *hist_length) {
	double **matrix;
	int nrows, ncols;
	long i;
	
	Read_Matrix(file_name, &matrix, &nrows, &ncols);
	*hist_length = nrows;
	*history = (event *) realloc(*history, nrows*sizeof(event));
	for (i = 0; i < nrows; ++i) {
		(*history)[i].state = (short) matrix[i][0];
		(*history)[i].time = (double) matrix[i][1];
		free(matrix[i]);
	}
	free(matrix);
}



event *Build_History (long chosen_cell, long *return_hist_len) {
	/* Returns the history of the chosen cell */
	
	long last_round, i, j, ancestor_cell, cur_cell, cur_hist_length, chosen_line, chosen_hist_len;
	char file_name[100], line[1001], *the_state, *the_time;
	short keep_reading;
	FILE *input_file;
	event *cur_history, *chosen_history;
	long  max_hist_len;  /* this is the maximal number of descendants initially allocated */
	long *file_pos;
	
	last_round = Round-1;
	max_hist_len = 1000;
	chosen_hist_len = 0;
	chosen_history = (event *) calloc(max_hist_len, sizeof(event));
	cur_history = (event *) calloc(1000, sizeof(event));
	chosen_line = chosen_cell;
	file_pos = (long *) calloc(last_round+1, sizeof(long));
	
	sprintf(file_name, "%s/history_file", Output_Dir); 
	input_file = fopen(file_name, "r");
	
	/* keep track of locations in history_file where history files for each round are */
	/* store these locations into file_pos array */
	while (fgets(line, 1000, input_file)!=NULL) {
		if (!strcmp(strtok(line, " \t"),"file")) {
			i = (long) atof(strtok(NULL, " \t"));
			file_pos[i] = ftell(input_file);
		}
	}
	
	for (i = last_round; i >= 0; --i) {
		/* jump to the position in history_file where the historical records 
		 for the present round are located */
		fseek(input_file, file_pos[i], SEEK_SET);
		
		cur_cell = -1;
		
		/* loop until correct cell is found, then read its history */
		while ((cur_cell != chosen_line) && (fgets(line, 1000, input_file)!=NULL)) {
			cur_cell = (long) atof(strtok(line, " \t"));
			ancestor_cell =  (long) atof(strtok(NULL, " \t"));
			keep_reading = 1;   
			cur_hist_length = 0;
			while (keep_reading) {     
				fgets(line, 1000, input_file);
				the_state = strtok(line, " \t");
				the_time = strtok(NULL, " \t");
				if (*the_state == '*')
					keep_reading = 0;
				else {
					cur_history[cur_hist_length].state = (long) atof(the_state);
					cur_history[cur_hist_length].time = (double) atof(the_time);
					++cur_hist_length;
				}
			}
		}
		
		/* update the chosen line cell so it traces backwards from chosen_cell */
		chosen_line = ancestor_cell; 
		
		/* increase max-size of current history if necessary */
		if (cur_hist_length+chosen_hist_len>max_hist_len) {
			chosen_history = realloc(chosen_history, 2*max_hist_len*sizeof(event));
			max_hist_len *= 2;
		}
		
		/* save current history in chosen history*/
		for (j = cur_hist_length-1; j >= 0; --j)
			chosen_history[chosen_hist_len++] = cur_history[j];
		
	}
	
	fclose(input_file);
	
	free(cur_history);
	free(file_pos);
	
	*return_hist_len = chosen_hist_len;
	
	return(chosen_history);
}



void Choose_History() {
	int i, chosen_cell;
	long chosen_hist_len, env_hist_length;
	FILE *file;
	char file_name[2000], hist_file_name[2000], env_file_name[2000];
	event *chosen_history, *env_history;
	double sim_length;
	
	
	env_history = (event *) calloc(1000, sizeof(event));
	
	sprintf(hist_file_name, "%s/chosen_histories", Output_Dir);
	sprintf(env_file_name, "%s/chosen_envs", Output_Dir);
	
	sprintf(file_name, "%s/env_file", Output_Dir);
	Read_History(file_name, &env_history, &env_hist_length);
	
	sim_length = env_history[env_hist_length-1].time;
	
	chosen_cell = Randomint(Pop_Sample);
	chosen_history = Build_History(chosen_cell, &chosen_hist_len);
	
	file = fopen(hist_file_name, "a");
	for (i = chosen_hist_len-1; i >= 0; --i)
		fprintf(file, "%ld %.4lf ", chosen_history[i].state, chosen_history[i].time);
	fprintf(file, "\n");
	fclose(file);
	
	file = fopen(env_file_name, "a"); 
	for (i = 0; i < env_hist_length; ++i)
		fprintf(file, "%ld %.4lf ", env_history[i].state, env_history[i].time);
	fprintf(file, "\n");
	fclose(file);
	
	free(chosen_history);
	free(env_history);
}


void Init_Population (void) {
	long i, j, k;
	
	UniqueID = 1;
	for (i = 0; i < NState; ++i)
		Pop[i] = 0;
	
	if (Record_History) {
		if (!Use_Initial_Pop) {                       /* use a single founding cell */
			Save_State[0] = Randomint(NState);
			Save_ID[0] = UniqueID++;
			NCell_Sample = 1;
		}
		else {
			k = 0;
			for (i = 0; i < NState; ++i)
				for (j = 0; j < Initial_Pop[i]; ++j) {      /* use several founding cells, as specified */
					Save_ID[k] = UniqueID++;
					Save_State[k++] = i;
				}
			NCell_Sample = k;
		}
	}
	else {
		if (!Use_Initial_Pop)  {                         /* use single founding cell */
			Pop[Randomint(NState)]++;
			NCell_Sample = 1;
		}
		else {                                           /* use several founding cells */
			if (Competition)
				for (j = 0; j < NStrains; ++j)
					Strain_Pop[j]=0;
			for (i = 0; i < NState; ++i) {
				Pop[i] = Initial_Pop[i];
				if (Competition)
					Strain_Pop[Strain_ID[i]] += Pop[i];
			}
			NCell_Sample = Get_Pop_Size();
		}
	}
	NPop = NCell_Sample;
}


int main(int argc, char *argv[]) {
	long i, j, k, *save_cell, run, save_pop_max, nextinct;
	FILE *history_file;
	double frac, save_sim_length, prob_ext;
	char hist_file_name[1000], env_file_name[1000];
	short save_record_history;
	
	Inits(argc, argv);
	
	sprintf(hist_file_name, "%s/history_file", Output_Dir);
	sprintf(env_file_name,"%s/env_file", Output_Dir);
	
    nextinct = 0;
	/* loop over independent runs */
	for(run=0; run<NRuns; ++run) {
		
		/* open history and environment files */
		History_File = fopen(hist_file_name, "w");
		Environment_File = fopen(env_file_name, "w");
		fprintf(Environment_File, "0 0.000\n");   /* record the start time as initial event */
		
		/* initialize time counters*/
		Ttot = Tlast = Tcycle = 0;
		Round = 1;
		
		/* initialize environment */
		if (Initial_Env == -1)
			Cur_Env = Randomint(NEnv);
		else
			Cur_Env = Initial_Env;
		if (NEnv>1)
			Tperiod = Sample_Distribution(Distributions[Cur_Env]);
		else
			Tperiod = Sim_Length+1;
		
		/* initialize population vector */
		Init_Population();
		
		/* pre-equilibration: process is not recorded for histories or otherwise */
		if (Pre_Equilibrate) {
			save_sim_length = Sim_Length;
			save_pop_max = Pop_Max;
			save_record_history = Record_History;
			Record_History=0;
			Pop_Max = NStrains*Pop_Max;
			for (i = 0; i < NState; ++i)
				Pop[i] = 0;
			for (i = 0; i < NCell_Sample; ++i)
				Pop[Save_State[i]]++;
			Sim_Length = Sample_Distribution(Pre_Equil_Dist);
			Calc_Rates(Rate_Matrices[Cur_Env], NState, Pop, Rates, &Tot_Rate);
			Setup_Probs(Rate_Matrices[Cur_Env], NState, Pop, Rates, Tot_Rate);
			while (Ttot < Sim_Length) {
				Simulate();
				if (Competition)
					Resample_Separately();
				else
					Resample();
			}
			printf("pre-equilibrium:\n");
			Print_State();
			Resample();
			Ttot = Tlast = 0;
			Sim_Length = save_sim_length;
			Pop_Max = save_pop_max;
			Record_History = save_record_history;
			printf("start:\n");
			Print_State();
		}
		
		/* output initialization lines to history file */
		if (Record_History) {
			fprintf(History_File, "file 0\n");
			for (i = 0; i < NCell_Sample; ++i) {
				fprintf(History_File, "%ld 0\n", i);
				fprintf(History_File, "0 0\n");
				/* this 0 0 guarantees that when history is later built from these files, 
				 it will include the start time of zero as its initial event */ 
				fprintf(History_File, "* *\n");
			}
		}
		
		Lyap = 0;
		Avg_pop = 0;
		Tcycle = 0;
		while ((NPop > 0) && (Ttot < Sim_Length) && (!Competition || No_Winner())) {
			
			/* re-sample population */
			if (Record_History) { 
				/* allocate memory for histories */
				Init_History();
				
				for (i = 0; i < NState; ++i)
					Pop[i] = 0;
				for (i = 0; i < NCell_Sample; ++i) {
					Cell_State[i] = Save_State[i];
					Cell_ID[i] = Save_ID[i];
					Pop[Save_State[i]]++;
					Ancestor[i].cell = -1;
					Ancestor[i].time = Ttot;
				}
				for (i = 0; i < NState; ++i) {
					Cell_List_Length[i] = Pop[i];
					k = 0;
					for (j = 0; j < NCell_Sample; ++j)
						if (Cell_State[j]==i)
							Cell_List[i][k++] = j;
				}
			}
			else if (Ttot > 0)
				Resample();
			
			/* calculates the total rate for each process (independent of population vector)
			 and for all processes (dependent on population vector) */
			Calc_Rates(Rate_Matrices[Cur_Env], NState, Pop, Rates, &Tot_Rate);
			
			/* setup cumulative probability densities of jumping "from" and "to" any state */
			Setup_Probs(Rate_Matrices[Cur_Env], NState, Pop, Rates, Tot_Rate);
			
            Print_State();
			
			Simulate();
			
			/* set population size for resampling... 
			 ... allows first round to have different initial pop size than next rounds */
			NCell_Sample = Pop_Sample;
			
            if (Tcycle >= Tperiod) {
                Tcycle = 0;
                /* record the state and time-stamp of environment before the change */
                fprintf(Environment_File, "%d %.4lf\n", Cur_Env, Ttot);
                
                /* change relevant variables/parameters to those of the next period */
                Cur_Env = Discrete_Sample(Env_Cume[Cur_Env], NEnv);
                Tperiod = Sample_Distribution(Distributions[Cur_Env]);
            }
            
			if (Record_History) {
				save_cell = ksub_random2(NCell_Sample, NPop);
				for (i = 0; i < NCell_Sample; ++i)
					save_cell[i] = Alive_Cells[save_cell[i]];
				free(Alive_Cells);
				NPop = NCell_Sample;
				Output_Histories(save_cell, NCell_Sample, Round); 
				for (i = 0; i < NCell_Sample; ++i) {
					Save_State[i] = Cell_State[save_cell[i]]; 
					Save_ID[i] = Cell_ID[save_cell[i]];
				}
				free(save_cell);
				Free_History();
			}
			
            if (NPop ==0) ++nextinct;
			++Round;
		}
		
		fclose(Environment_File);
		fclose(History_File);
		
		Print_State();
		
		/* output all simulation parameters to the file histories/parameters */
		Output_Parameters();
		
		if (Record_History && !Record_All_Cells)
			Choose_History();
		
		if (Competition) {
			printf("run %ld: %d\n",run, Who_Won());
			fflush(stdout);
		}
        
	}
    
   // prob_ext =(double) nextinct / NRuns;
   // printf("extinct: %ld\t %ld\t %lf\t %lf\n", nextinct, NRuns,prob_ext , sqrt(prob_ext*(1-prob_ext)/NRuns));
	
}
